@@ -4,6 +4,7 @@
 import argparse
 import configparser
 import fnmatch
+import io
 import os
 import re
 import shutil
@@ -12,34 +13,63 @@ import typing as t
 import zipfile
 from datetime import datetime
 
-__version__ = "0.2.10"
+
+# Current version number
+__version__ = "0.3.0"
 
 
-class JMLHelpFormatter(argparse.HelpFormatter):
-    """Custom HelpFormatter to replace spaces with underscores in
-    metavars.
-    """
+# Some constants
+RE_VERSION = re.compile(r"^\d+$")
+RE_VERSION2 = re.compile(r"^([!<>=]{0,2})(\d+)$")
+RE_REPLACE = re.compile(r"^(?<!\\)/(.*?)(?<!\\)/(.*)(?<!\\)/$")
 
-    def __init__(self, prog, indent_increment=2, max_help_position=24, width=None):
-        super().__init__(prog, indent_increment=2, max_help_position=24, width=None)
+ML_INT = -1
 
-    def _metavar_formatter(self, action, default_metavar):
-        sformat = super()._metavar_formatter(action, default_metavar)
+CONFIG_FILE = ".jml"
+CONFIG_SECTION = "settings"
+DEFAULT_CONFIG = {
+    "name": "",
+    "output dir": "",
+    "task open": "/*aufg*",
+    "task close": "*aufg*/",
+    "task comment prefix": "",
+    "solution open": "//ml*",
+    "solution close": "//*ml",
+    "solution comment prefix": "",
+    "solution suffix": "ML",
+    "name format": "{project}_{version}",
+    "include": "*.java",
+    "+include": "",
+    "-include": "",
+    "exclude": "*.class,*.ctxt,.DS_Store,Thumbs.db,*.iml,.vscode,.eclipse",
+    "+exclude": "",
+    "-exclude": "",
+    "create zip": "no",
+    "create zip only": "no",
+    "create zip dir": "",
+    "encoding": "utf-8",
+    "additional files": "",
+    "+additional files": "",
+    "-additional files": "",
+    "project root": "",
+    "clear": "yes",
+    "delete solution": "no",
+    "keep empty files": "yes",
+    "keep empty folders": "yes",
+}
 
-        def new_format(tuple_size):
-            t = sformat(tuple_size)
-            return tuple(r.replace(" ", "_") if r else r for r in t)
-
-        return new_format
+DEBUG_FLAG = False
+DRY_RUN = False
 
 
+# Initialize argument parser
 parser = argparse.ArgumentParser(
     prog="jml",
     description="Generiert aus einem Basisprojekt mehrere Projektversionen.",
     add_help=False,
-    formatter_class=JMLHelpFormatter,
 )
 
+# version and help
 parser.add_argument(
     "-h",
     "--help",
@@ -53,10 +83,10 @@ parser.add_argument(
     help="Programmversion anzeigen.",
 )
 
-# Argumente
+# arguments
 parser.add_argument("srcdir", metavar="IN", help="Pfad des Basisprojektes")
 
-# Optionen
+# options
 parser.add_argument(
     "-o",
     "--outdir",
@@ -75,6 +105,7 @@ parser.add_argument(
     "-nf",
     "--name-format",
     dest="name format",
+    metavar="FORMAT",
     action="store",
     help="Format für die Namen der Projektversionen. Angabe als Python-Formatstring. Kann die Variablen {project}, {version} und {date} enthalten. Standard: {project}_{version}",
 )
@@ -82,6 +113,7 @@ parser.add_argument(
     "-mls",
     "--ml-suffix",
     dest="solution suffix",
+    metavar="SUFFIX",
     action="store",
     help="Suffix für die Musterlösung. Standard: ML",
 )
@@ -89,6 +121,7 @@ parser.add_argument(
     "-to",
     "--tag-open",
     dest="task open",
+    metavar="TAG",
     action="store",
     help="Öffnende Aufgaben-Markierung.Standard: /*aufg*",
 )
@@ -96,6 +129,7 @@ parser.add_argument(
     "-tc",
     "--tag-close",
     dest="task close",
+    metavar="TAG",
     action="store",
     help="Schließende Aufgaben-Markierung. Standard: *aufg*/",
 )
@@ -103,6 +137,7 @@ parser.add_argument(
     "-mlo",
     "--ml-open",
     dest="solution open",
+    metavar="TAG",
     action="store",
     help="Öffnende Lösungs-Markierung. Standard: //ml*",
 )
@@ -110,6 +145,7 @@ parser.add_argument(
     "-mlc",
     "--ml-close",
     dest="solution close",
+    metavar="TAG",
     action="store",
     help="Schließende Lösungs-Markierung. Standard: //*ml",
 )
@@ -133,13 +169,16 @@ parser.add_argument(
     "-v",
     "--versions",
     dest="versions",
+    metavar="VER",
     action="extend",
+    nargs="+",
     type=str,
     help="Liste von Versionen, die erstellt werden sollen.",
 )
 parser.add_argument(
     "--project-root",
     dest="project root",
+    metavar="PATH",
     action="store",
     help="Bestimmt ein Verzeichnis, das als Wurzelverzeichnis für mehrere Projekte dient. Wenn dies ein Prefix von IN ist, dann reflektiert die Ordnerstruktur in OUT die des Wurzelverzeichnisses. Für IN=/projects/foo/bar/my-project OUT=/projects/out project-root=/projects werden die Projektversionen dann in /projects/out/foo/bar erstellt.",
 )
@@ -184,77 +223,46 @@ parser.add_argument(
     action="store_true",
     help="Zeigt Informationen zur Fehlersuche an.",
 )
+parser.add_argument(
+    "--dry-run",
+    dest="dry_run",
+    action="store_true",
+    help="Führt keine Operationen aus, sondern zeigt nur an, welche Änderungen vorgenommen würden. Impliziert --debug.",
+)
 
 
-RE_VERSION = re.compile(r"^\d+$")
-RE_VERSION2 = re.compile(r"^([!<>=]{0,2})(\d+)$")
-RE_REPLACE = re.compile(r"^(?<!\\)/(.*?)(?<!\\)/(.*)(?<!\\)/$")
-
-CONFIG_FILE = ".jml"
-CONFIG_SECTION = "settings"
-DEFAULT_CONFIG = {
-    "name": "",
-    "output dir": "",
-    "task open": "/*aufg*",
-    "task close": "*aufg*/",
-    "task comment prefix": "",
-    "solution open": "//ml*",
-    "solution close": "//*ml",
-    "solution comment prefix": "",
-    "solution suffix": "ML",
-    "name format": "{project}_{version}",
-    "include": "*.java",
-    "+include": "",
-    "-include": "",
-    "exclude": "*.class,*.ctxt,.DS_Store,Thumbs.db,*.iml,.vscode,.eclipse",
-    "+exclude": "",
-    "-exclude": "",
-    "create zip": "no",
-    "create zip only": "no",
-    "create zip dir": "",
-    "encoding": "utf-8",
-    "additional files": "",
-    "+additional files": "",
-    "-additional files": "",
-    "project root": "",
-    "clear": "yes",
-    "delete solution": "no",
-    "keep empty files": "yes",
-    "keep empty folders": "yes",
-}
-
-DEBUG_FLAG = False
-
-
+# simple debug function
 def debug(msg: str, indent: int = 0) -> None:
     """Shows a debug message if debugging is enabled."""
-    if globals()["DEBUG_FLAG"]:
+    if DEBUG_FLAG:
         if indent:
             print(indent * "  " + msg)
         else:
             print(msg)
 
 
+# main function
 def main() -> None:
     args = parser.parse_args()
 
-    debug_flag = args.debug
-    globals()["DEBUG_FLAG"] = debug_flag
+    debug_flag = bool(args.debug)
+    globals()["DRY_RUN"] = args.dry_run
+    globals()["DEBUG_FLAG"] = debug_flag or args.dry_run
 
     # check srcdir
-    srcdir = args.srcdir = resolve_path(args.srcdir)
+    srcdir = resolve_path(args.srcdir)
     if not os.path.isdir(srcdir):
         print(
             f"Quellverzeichnis <{srcdir}> ist nicht vorhanden! Bitte wähle ein gültiges Projektverzeichnis."
         )
         quit()
 
-    # resolve final rootdir (requires to get final project root from config)
+    # resolve rootdir (requires to read project config to get final project root)
     project_root = vars(args)["project root"]
     if project_root:
         project_root = resolve_path(project_root)
 
-    # read project config first, to get final root dir
+    ### read project config, to get final rootdir
     proj_config = None
     proj_config_file = os.path.join(srcdir, CONFIG_FILE)
     if os.path.exists(proj_config_file):
@@ -294,7 +302,7 @@ def main() -> None:
             del settings[k]
 
     # add cli arguments
-    merge_configs(
+    args_to_config(
         settings,
         args,
         flags={
@@ -310,7 +318,7 @@ def main() -> None:
         settings["name"] = os.path.basename(srcdir)
 
     # show config for debugging
-    if debug_flag:
+    if DEBUG_FLAG:
         debug("config loaded:")
         for k, v in sorted(settings.items()):
             debug(f"{k} = {v}", 1)
@@ -363,7 +371,7 @@ def main() -> None:
     debug(f"  to <{outdir}>", 1)
 
     debug("Creating solution version:")
-    versions = create_version(-1, settings)
+    versions = create_version(ML_INT, settings)
     if max(versions) > 0:
         versions = {v + 1 for v in range(max(versions))}
 
@@ -387,7 +395,7 @@ def create_version(version: int, settings: configparser.SectionProxy) -> t.Set[i
     srcdir = settings["srcdir"]
     outdir = settings["output dir"]
     name = settings["name"]
-    is_ml = version < 0
+    is_ml = version == ML_INT
 
     versions = set()
 
@@ -409,12 +417,14 @@ def create_version(version: int, settings: configparser.SectionProxy) -> t.Set[i
     # prepare output folders
     if os.path.isdir(outdir):
         if settings.getboolean("clear"):
-            shutil.rmtree(outdir)
+            if not DRY_RUN:
+                shutil.rmtree(outdir)
             debug(f"removed target directory <{outdir}>", 1)
         else:
             debug(f"using existing target directory at <{outdir}>", 1)
     if not os.path.isdir(outdir):
-        os.makedirs(outdir)
+        if not DRY_RUN:
+            os.makedirs(outdir)
         debug(f"created target directory <{outdir}>", 1)
 
     # extract some config options to local scope
@@ -424,38 +434,11 @@ def create_version(version: int, settings: configparser.SectionProxy) -> t.Set[i
 
     tag_open = settings["task open"]
     tag_close = settings["task close"]
-    task_replace = lambda l: l  # noqa: E731
-    if settings["task comment prefix"]:
-        m = RE_REPLACE.match(settings["task comment prefix"])
-        if m:
-            tpat, trepl = m.group(1), m.group(2)
-            tpat, trepl = re.compile(tpat.replace("\\/", "/")), trepl.replace("\\/", "/")
-
-            def task_replace(line):
-                return re.sub(tpat, trepl, line)
-        else:
-            tpat = re.compile(f"^(\\s*)({settings['task comment prefix']})")
-
-            def task_replace(line):
-                # return re.sub(tpat, lambda m: f"{m.group(1)}{' '*len(m.group(2))}", line)
-                return re.sub(tpat, "\\1", line)
 
     ml_open = settings["solution open"]
     ml_close = settings["solution close"]
-    ml_replace = lambda l: l  # noqa: E731
-    if settings["solution comment prefix"]:
-        if settings["solution comment prefix"].startswith("\\"):
-            spat, srepl, _ = re.split(r"(?<!\\)\/", settings["solution comment prefix"][1:], 2)
-            spat, srepl = re.compile(spat.replace("\\/", "/")), srepl.replace("\\/", "/")
 
-            def ml_replace(line):
-                return re.sub(spat, srepl, line)
-        else:
-            spat = re.compile(f"^(\\s*)({settings['solution comment prefix']})")
-
-            def ml_replace(line):
-                # return re.sub(spat, lambda m: f"{m.group(1)}{' '*len(m.group(2))}", line)
-                return re.sub(tpat, "\\1", line)
+    transform_func = create_transform(version, settings)
 
     keep_empty_files = settings.getboolean("keep empty files")
 
@@ -467,7 +450,8 @@ def create_version(version: int, settings: configparser.SectionProxy) -> t.Set[i
         subpath = root[strippos:]
         outroot = os.path.join(outdir, subpath)
 
-        os.makedirs(outroot, exist_ok=True)
+        if not DRY_RUN:
+            os.makedirs(outroot, exist_ok=True)
 
         for file in files:
             fullpath = os.path.join(root, file)
@@ -481,7 +465,7 @@ def create_version(version: int, settings: configparser.SectionProxy) -> t.Set[i
             elif match_patterns(file, include):
                 is_empty = True
                 with open(fullpath, "r", encoding=encoding) as inf:
-                    with open(fulloutpath, "w", encoding=encoding) as outf:
+                    with open_file(fulloutpath, encoding=encoding) as outf:
                         skip = False
                         transform = None
                         line = inf.readline()
@@ -496,7 +480,7 @@ def create_version(version: int, settings: configparser.SectionProxy) -> t.Set[i
                                 transform = None
                             elif lline.startswith(ml_open):
                                 skip = not is_ml
-                                transform = ml_replace
+                                transform = transform_func
                             elif lline.startswith(tag_open):
                                 parts = lline.split()
                                 if len(parts) > 1:
@@ -509,7 +493,7 @@ def create_version(version: int, settings: configparser.SectionProxy) -> t.Set[i
                                         skip = not test_version(version, parts[1])
                                 else:
                                     skip = is_ml
-                                transform = task_replace
+                                transform = transform_func
                             elif skip:
                                 pass
                             else:
@@ -533,16 +517,19 @@ def create_version(version: int, settings: configparser.SectionProxy) -> t.Set[i
         if file:
             if os.path.isfile(file):
                 fulloutpath = os.path.join(outdir, os.path.basename(file))
-                shutil.copy(file, fulloutpath)
+                if not DRY_RUN:
+                    shutil.copy(file, fulloutpath)
                 debug(f"{file:>32} -> {fulloutpath}", 2)
 
     if is_ml and settings.getboolean("delete solution"):
-        shutil.rmtree(outdir)
+        if not DRY_RUN:
+            shutil.rmtree(outdir)
         debug("removed compiled ml directory", 1)
     elif settings.getboolean("create zip") or settings.getboolean("create zip only"):
         create_zip(outdir, settings)
         if settings.getboolean("create zip only"):
-            shutil.rmtree(outdir)
+            if not DRY_RUN:
+                shutil.rmtree(outdir)
             debug("removed compiled version directory", 1)
 
     if not versions:
@@ -558,26 +545,27 @@ def create_zip(path: str, settings: configparser.SectionProxy) -> None:
         # prepare output directory
         if settings["create zip dir"]:
             dir = resolve_path(settings["create zip dir"])
-        if not os.path.exists(dir):
+        if not os.path.exists(dir) and not DRY_RUN:
             os.makedirs(dir)
 
         # prepare output filename
         filename = f"{filename}.zip"
         outfile = os.path.join(dir, filename)
-        if os.path.isfile(filename):
+        if os.path.isfile(outfile) and not DRY_RUN:
             os.remove(outfile)
-        elif os.path.isdir(filename):
+        elif os.path.isdir(outfile):
             debug(f"directory found at <{outfile}>! unable to create zip file.", 1)
             return
 
         # create zip file
-        with zipfile.ZipFile(outfile, "w", zipfile.ZIP_DEFLATED) as zipf:
-            for root, dirs, files in os.walk(path):
-                for file in files:
-                    filepath = os.path.join(root, file)
-                    relpath = os.path.relpath(filepath, start=path)
-                    zipf.write(filepath, arcname=relpath)
-            debug(f"created zip file at <{outfile}>", 1)
+        if not DRY_RUN:
+            with zipfile.ZipFile(outfile, "w", zipfile.ZIP_DEFLATED) as zipf:
+                for root, dirs, files in os.walk(path):
+                    for file in files:
+                        filepath = os.path.join(root, file)
+                        relpath = os.path.relpath(filepath, start=path)
+                        zipf.write(filepath, arcname=relpath)
+        debug(f"created zip file at <{outfile}>", 1)
 
 
 def resolve_path(path: str, base: str = None) -> str:
@@ -635,7 +623,7 @@ def resolve_config(config: configparser.ConfigParser, base: str = None) -> None:
             config.set(CONFIG_SECTION, opt, ",".join(value))
 
 
-def merge_configs(
+def args_to_config(
     config: configparser.SectionProxy,
     args: argparse.Namespace,
     flags: t.Dict[str, bool] = dict(),
@@ -651,7 +639,7 @@ def merge_configs(
     destination and the values the boolean that is stored (e.g. True for store_true).
     This is necessary because boolean flags are always set and will override
     any value already set in the config object. By providing the flags dict,
-    the option is only overriden, if the flag was supplied with the arguments.
+    the option is only overridden, if the flag was supplied with the arguments.
     """
     for k, v in vars(args).items():
         if v is not None:
@@ -665,6 +653,38 @@ def merge_configs(
                     config[k] = str(v)
 
 
+def open_file(path: str, mode: str = "w", encoding: str = "utf-8") -> t.IO[t.Any]:
+    if mode[0] == "w" and DRY_RUN:
+        return io.StringIO()
+    else:
+        return open(path, mode, encoding=encoding)
+
+
+def create_transform(version: int, settings: configparser.ConfigParser) -> t.Callable:
+    transform = lambda line: line  # noqa: E731
+
+    prefix = settings["task comment prefix"]
+    if version == ML_INT:
+        prefix = settings["solution comment prefix"]
+
+    if prefix:
+        m = RE_REPLACE.match(prefix)
+        if m:
+            pat, repl = m.group(1).replace("\\/", "/"), m.group(2).replace("\\/", "/")
+            pat, repl = re.compile(pat), repl
+
+            def transform(line):
+                return re.sub(pat, repl, line)
+
+        else:
+            pat = re.compile(f"^(\\s*)({prefix})")
+
+            def transform(line):
+                return re.sub(pat, "\\1", line)
+
+    return transform
+
+
 def match_patterns(filename: str, patterns: t.List[str]) -> bool:
     """Matches a filename against a list of UNIX-like filename patterns.
     True is returned, if filename matches at least one pattern.
